#![feature(const_mut_refs)]

use std::{marker::PhantomData, sync::Arc, any::{TypeId, Any}, ops::{Deref, DerefMut}, fmt::Debug, borrow::Borrow};

use ahash::AHashMap;
use futures::executor::block_on;
use tokio::{sync::mpsc::{self, error::TryRecvError}, task};
use dyn_clone::{DynClone, clone_box};
// use futures::executor::block_on;
use parking_lot::{Mutex, RwLock, RwLockReadGuard, RwLockWriteGuard};

#[cfg(test)]
mod test;

#[derive(Debug)]
pub struct Context {
    data: Arc<RwLock<ContextData>>
}

impl Context {
    pub fn new() -> Self {
        let data = ContextData::new();

        data.event_manager.start();
        
        Self {
            data: Arc::new(RwLock::new(data))
        }
    }

    pub fn add_dependency(&mut self, dependency: Dependency) {
        self.get_or_init_system(&dependency);
    }

    pub fn remove_dependency(&mut self, dependency: Dependency) {
        todo!("Not yet implemented, for now everything is only additive")
    }
    
    fn get_or_init_system(&mut self, system: &Dependency) -> usize {
        match self.get_system_index(system) {
            Some(index) => index,
            None => self.init_system(system),
        }
    }

    fn init_system(&mut self, system: &Dependency) -> usize {
        let dependencies = system.dependencies;
        let mut dependency_ids = vec![];
        for dependency in dependencies {
            dependency_ids.push((dependency.id.into(), self.get_or_init_system(dependency)));
        }

        let event_sender = self.data.read().event_manager.get_sender();
        
        let context_handle = ContextHandle {
            context: self.data.clone(),
            dependency_ids: dependency_ids,
            event_sender: event_sender,
        };

        let builder = (system.builder)();
        let system_handle = builder.create(context_handle);
        
        let mut data = self.data.write();
        let index = data.systems.len();
        
        let event_handlers = system.event_handlers;
        for event_handler in event_handlers {
            let boxed_handler: Box<dyn EventHandlerWrapper> = clone_box(event_handler.to_owned());
            let event_handler_handle = EventHandlerHandle::new(system_handle.system.clone(), boxed_handler);
            data.event_manager.register_handler(event_handler_handle);
        }
        
        data.systems.push(system_handle);

        return index;
    }

    fn get_system_index(&self, dependency: &Dependency) -> Option<usize> {
        let id = dependency.id.into();
        let systems = &self.data.read().systems;
        for i in 0..systems.len() {
            if systems[i].id == id {
                return Some(i);
            }
        }
        None
    }
}

#[derive(Debug)]
struct ContextData {
    systems: Vec<SystemHandle>,
    event_manager: EventManager
}

impl ContextData {
    fn new() -> Self {
        Self {
            systems: vec![],
            event_manager: EventManager::new()
        }
    }
}

#[derive(Clone)]
pub struct ContextHandle {
    context: Arc<RwLock<ContextData>>,
    dependency_ids: Vec<(TypeId, usize)>,
    event_sender: mpsc::UnboundedSender<Event>
}

impl ContextHandle {
    #[tracing::instrument]
    pub fn obtain<S: System>(&self) -> SystemRef<S> {
        let index = self.dependency_index::<S>();
        // let system = self.context.blocking_read().systems[index].system.clone();
        let ctx_data = self.context.read();
        let system = ctx_data.systems[index].system.clone();

        SystemRef::new(system)
    }

    pub fn raise<T: 'static + Send + Sync>(&self, event: T) {
        let event = Event::new(event);
        self.event_sender.send(event).expect("ContextHandle failed to transmit event");
    }
    
    fn dependency_index<S: System>(&self) -> usize {
        let id = TypeId::of::<S>();
        for dependency in &self.dependency_ids {
            if dependency.0 == id {
                return dependency.1
            }
        }
        panic!("Requested invalid dependency");
    }
}

impl Debug for ContextHandle {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("ContextHandle").field("dependency_ids", &self.dependency_ids).finish()
    }
}

pub struct SystemRef<S: System> {
    lock: Arc<RwLock<dyn Any>>,
    marker: PhantomData<fn(S)>
}

impl<S: System> SystemRef<S> {
    fn new(lock: Arc<RwLock<dyn Any>>) -> Self {
        Self {
            lock: lock,
            marker: PhantomData
        }
    }

    pub fn get(&self) -> SystemReadGuard<S> {
        SystemReadGuard::new(&self.lock)
    }

    pub fn get_mut(&mut self) -> SystemWriteGuard<S> {
        SystemWriteGuard::new(&self.lock)
    }
}

pub struct SystemReadGuard<'a, S: System> {
    guard: RwLockReadGuard<'a, dyn Any>,
    marker: PhantomData<fn(S)>
}

impl<'a, S: System> SystemReadGuard<'a, S> {
    fn new(lock: &'a Arc<RwLock<dyn Any>>) -> Self {
        Self {
            guard: lock.read(),
            marker: PhantomData
        }
    }
}

impl<'a, S: System> Deref for SystemReadGuard<'a, S> {
    type Target = S;

    fn deref(&self) -> &Self::Target {
        self.guard.downcast_ref().expect("SystemReadGuard contained incorrect system")
    }
}

pub struct SystemWriteGuard<'a, S: System> {
    guard: RwLockWriteGuard<'a, dyn Any>,
    marker: PhantomData<fn(S)>
}

impl<'a, S: System> SystemWriteGuard<'a, S> {
    fn new(lock: &'a Arc<RwLock<dyn Any>>) -> Self {
        Self {
            guard: lock.write(),
            marker: PhantomData
        }
    }
}

impl<'a, S: System> Deref for SystemWriteGuard<'a, S> {
    type Target = S;

    fn deref(&self) -> &Self::Target {
        self.guard.downcast_ref().expect("SystemReadGuard contained incorrect system")
    }
}

impl<'a, S: System> DerefMut for SystemWriteGuard<'a, S> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        self.guard.downcast_mut().expect("SystemWriteGuard contained incorrect system")
    }
}

pub trait System: 'static + Sized + Send + Sync {
    const DEPENDENCIES: Dependencies;
    const EVENT_HANDLERS: EventHandlers;

    fn new(ctx: ContextHandle) -> Self;

    fn start(&mut self);
}

#[derive(Debug)]
struct SystemHandle {
    system: Arc<RwLock<dyn Any + Send + Sync>>,
    // context_handle: ContextHandle,
    id: TypeId
}

pub type Dependencies = &'static [&'static Dependency];

pub struct Dependency {
    builder: fn() -> Box<dyn SystemBuilder>,
    dependencies: Dependencies,
    event_handlers: EventHandlers,
    id: ConstTypeId
}

impl Dependency {
    pub const fn new<S: System + Send + Sync>() -> Self {
        Self {
            builder: TypedSystemBuilder::<S>::get_builder,
            dependencies: S::DEPENDENCIES,
            event_handlers: S::EVENT_HANDLERS,
            id: ConstTypeId::of::<S>()
        }
    }
}

struct EventHandlerHandle {
    system: Arc<RwLock<dyn Any + Send + Sync>>,
    id: TypeId,
    handler: Box<dyn EventHandlerWrapper>
}

impl EventHandlerHandle {
    fn new(system: Arc<RwLock<dyn Any + Send + Sync>>, handler: Box<dyn EventHandlerWrapper>) -> Self {
        Self {
            system: system,
            id: handler.get_event_type(),
            handler: handler,
        }
    }
}

impl Debug for EventHandlerHandle {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("EventHandlerHandle").field("system", &self.system).field("id", &self.id).finish()
    }
}

pub type EventHandlers = &'static [&'static dyn EventHandlerWrapper];

pub trait EventHandlerWrapper: Send + Sync + DynClone {
    // fn create(&self) -> EventHandlerHandle;
    fn get_event_type(&self) -> TypeId;
    
    fn handle(&self, system: &Arc<RwLock<dyn Any + Send + Sync>>, event: &dyn Any);
}

#[derive(Copy)]
pub struct EventHandler<S: System, T: 'static> {
    handler: fn(&mut S, &T)
}

impl<S: System, T> EventHandler<S, T> {
    pub const fn new(handler: fn(&mut S, &T)) -> Self {
        Self {
            handler,
        }
    }
}

impl<S: System, T> EventHandlerWrapper for EventHandler<S, T> {
    fn get_event_type(&self) -> TypeId {
        TypeId::of::<T>()
    }

    fn handle(&self, system: &Arc<RwLock<dyn Any + Send + Sync>>, event: &dyn Any) {
        let mut system_guard = system.write();
        let system = system_guard.downcast_mut::<S>().expect("EventHandler was given incorrect system");
        let event = event.downcast_ref::<T>().expect("EventHandler was given incorrect event");
        (self.handler)(system, event);
    }
}

impl<S: System, T> Clone for EventHandler<S, T> {
    fn clone(&self) -> Self {
        Self { handler: self.handler.clone() }
    }
}

trait SystemBuilder: Send {
    fn create(&self, ctx: ContextHandle) -> SystemHandle;
}

struct TypedSystemBuilder<S: System> {
    marker: PhantomData<fn(S)>
}

impl<S: System + Send + Sync> TypedSystemBuilder<S> {
    fn get_builder() -> Box<dyn SystemBuilder> {
        Box::new(Self {
            marker: PhantomData
        })
    }
}

impl<S: System + Send + Sync> SystemBuilder for TypedSystemBuilder<S> {
    fn create(&self, ctx: ContextHandle) -> SystemHandle {
        let mut system = S::new(ctx.clone());

        system.start();

        SystemHandle {
            system: Arc::new(RwLock::new(system)),
            // context_handle: ctx,
            id: TypeId::of::<S>()
        }
    }
}

#[derive(Clone, Copy, Debug)]
struct ConstTypeId {
    id: fn() -> TypeId
}

impl ConstTypeId {
    pub const fn of<T: 'static>() -> Self {
        Self {
            id: TypeId::of::<T>
        }
    }
}

impl From<ConstTypeId> for TypeId {
    fn from(value: ConstTypeId) -> Self {
        (value.id)()
    }
}

#[derive(Debug)]
struct EventResolver {
    handlers: Vec<EventHandlerHandle>,
    // ordered: Vec<EventHandlerHandle>
}

impl EventResolver {
    fn new() -> Self {
        Self {
            handlers: vec![],
            // ordered: vec![],
        }
    } 

    fn register_handler(&mut self, handler: EventHandlerHandle) {
        self.handlers.push(handler);
        self.rebuild_order();
    }

    fn rebuild_order(&mut self) {

    }

    #[tracing::instrument]
    fn resolve_event(&self, event: Event) {
        // TODO: Dependency ordering and multithreading
        for handler in &self.handlers {
            let event_any = event.value.as_ref();
            handler.handler.handle(&handler.system, event_any);
        }
    } 
}

#[derive(Debug)]
struct EventManager {
    data: Arc<RwLock<EventManagerData>>
}

impl EventManager {
    fn new() -> Self {
        Self {
            data: Arc::new(RwLock::new(EventManagerData::new()))
        }
    }

    fn start(&self) {
        let data_handle = self.data.clone();
        task::spawn(async {
            let data = data_handle;
            loop {
                let lock = data.read();
                lock.handle_events();
            }
        });
    }
    
    fn get_sender(&self) -> mpsc::UnboundedSender<Event> {
        self.data.read().sender.clone()
    }

    fn register_handler(&mut self, handler: EventHandlerHandle) {
        let data = self.data.read();
        if !data.handlers.read().contains_key(&handler.id) {
            data.handlers.write().insert(handler.id, EventResolver::new());
        }

        match data.handlers.write().get_mut(&handler.id) {
            Some(resolver) => resolver.register_handler(handler),
            None => panic!("EventResolver disappeared while registering handler"),
        };
    }
}

#[derive(Debug)]
struct EventManagerData {
    sender: mpsc::UnboundedSender<Event>,
    receiver: Arc<Mutex<mpsc::UnboundedReceiver<Event>>>,

    handlers: Arc<RwLock<AHashMap<TypeId, EventResolver>>>
}

impl EventManagerData {
    fn new() -> Self {
        let (tx, rx) = mpsc::unbounded_channel();
        
        Self {
            sender: tx,
            receiver: Arc::new(Mutex::new(rx)),
            handlers: Arc::new(RwLock::new(AHashMap::new()))
        }
    }

    #[tracing::instrument]
    fn handle_events(&self) {
        let mut receiver = self.receiver.lock();
        let event = receiver.try_recv();
        match event {
            Ok(event) => {
                self.handle_event(event);
            },
            Err(err) => match err {
                TryRecvError::Empty => (),
                TryRecvError::Disconnected => {
                    panic!("Event channel closed");
                },
            },
        }
        // match block_on(receiver.recv()) {
        //     Some(event) => self.handle_event(event),
        //     None => panic!("EventManager channel has been closed"),
        // }
    }

    #[tracing::instrument]
    fn handle_event(&self, event: Event) {
        let handlers = self.handlers.read();
        let handler = handlers.get(&event.id);
        match handler {
            Some(handler) => {
                handler.resolve_event(event);
            },
            None => (),
        }
        // if let Some(resolver) = self.handlers.read().get(&event.id) {
        //     resolver.resolve_event(event);
        // }
    }
}

#[derive(Debug)]
struct Event {
    id: TypeId,
    value: Box<dyn Any + Send + Sync>
}

impl Event {
    pub fn new<S: 'static + Send + Sync>(value: S) -> Self {
        Self {
            id: TypeId::of::<S>(),
            value: Box::new(value)
        }
    }
}